
/*****************************************************************
 * HMMER - Biological sequence analysis with profile HMMs
 * Copyright (C) 1992-1999 Washington University School of Medicine
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 *****************************************************************/

/* marwis_func.c
 * 
 * Functions written by Markus Wistrand to use with the HMMER package
 * 
 */

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <float.h>  /* marwis */
#include "marwis_func.h"

#include "structs.h"
#include "config.h"
#include "squid.h"
#include "funcs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif


/* Function: ReadWeights()
 *
 * Purpose:  Use external weights given in a file
 *           
 * Args:     aseq        - array of (0..nseq-1) aligned sequences
 *           nseq        - number of seqs in alignment  
 *           alen        - length of alignment
 *           wgt         - allocated [0..nseq-1] array of weights to be returned
 *           
 * Return:   (void)
 *           wgt is filled in.
 */
void
ReadWeights(char **sqname, int nseq, float *wgt, char *wfile)
{
    FILE            *fp;
    int             i,j;
    char            *sptr;
    int             n; 
    int             nbytes;
    int             numseq; // Number of seqs in weightfile
    int             nline; 
    char            *s;
    int             *beginIdx, *beginWgt;
    int             *endIdx, *endWgt;
    int             wgtLth, idxLth; 
    char            **Wsqname;
    float           *Wwgt; 
    char            *tmpWwgt;
    int             first;
    
    if ((fp = fopen(wfile, "r")) == NULL)
	Die("Failed to open HMMER weights file %s\n", wfile);

       if (fseek(fp,0,2) != 0){
                (void)fprintf(stderr, "ERROR: fseek failed\n");
                (void)exit(EXIT_FAILURE);
        }
        nbytes = ftell(fp);

        if (fseek(fp,0,0) != 0){
                (void)fprintf(stderr, "ERROR: fseek failed\n");
                (void)exit(EXIT_FAILURE);
        }  

	sptr = MallocOrDie (sizeof(char) * (nbytes));
	
	numseq = 0;
	fread(sptr,sizeof(char), nbytes, fp);
	sptr[nbytes] = '\0';     /* adding trailing \0 */

	beginIdx = MallocOrDie (sizeof(int) * nseq);
	endIdx = MallocOrDie (sizeof(int) * nseq);
	beginWgt = MallocOrDie (sizeof(int) * nseq);
	endWgt = MallocOrDie (sizeof(int) * nseq);
	Wwgt = MallocOrDie (sizeof(float) * nseq);
	tmpWwgt = MallocOrDie (sizeof(char) * 8);
	Wsqname = (char **) MallocOrDie (sizeof(char *) * nseq);
       
        for (i =0;i < nbytes;i++){
	    
	    if (sptr[i] == '\n') {
		if (sptr[i+1] == '%') {
		    first = i+1;
		}
	    }
	}
	

	nline = 1; /* We have a newline, in fact the last comment line */	
        for (i = first;i < nbytes;i++){
	    /* If newline, we note this and continue. If space, we continue */
	    if (sptr[i] == '\n')
		nline = 1;
	    else if (sptr[i] == ' ') {}

	    /* Lines we aren't interested in (Comments */
	    else if ((sptr[i] == '%') && nline == 1) {
		while (sptr[i] != '\n') {
		    i++;
		}
	    } 
	    /* Lines of interest to us. We know they start with  'index', followed by  'weight' */
	    else {
		if (nline) { /* newline means index is coming */
		    beginIdx[numseq] = i;
		    nline = 0;
		    while (sptr[i] != '\t') { 
			i++;
		    }
		    endIdx[numseq] = i;
		}
		else { /* not newline means weight is coming */
		    beginWgt[numseq] = i;
		    while (sptr[i] != '\n') {
			i++;
		    }
		    endWgt[numseq] = i;
		    nline = 1;	
		    numseq++;
		}
	    }
	}
	if (numseq != nseq) {
	  Die("Sequence weighting file has %i weights but there is %i sequences in the alignment", numseq, nseq);
	}

	/* Get the index and weights from the string and copy to tmp arrays */
	for (i=0; i<numseq; i++) {
	    wgtLth = endWgt[i] - beginWgt[i];
	    idxLth = endIdx[i] - beginIdx[i];
	    Wsqname[i] = (char *) MallocOrDie (sizeof(char) * (idxLth+1));

	    /* sre_strncopy adds trailing \0 to string
               Eq. to strncpy + Wsqname[i][idxLth]= '\0'; */ 
	    sre_strncpy(Wsqname[i], &sptr[beginIdx[i]], idxLth);
	    sre_strncpy(tmpWwgt, &sptr[beginWgt[i]], wgtLth);
	    Wwgt[i] = atof(tmpWwgt);
	}

	/* Now get the new weights into the Hmmer weights */
	for (i=0; i<numseq; i++) {
	    for (j=0; j<numseq; j++) {
		if (strcmp(Wsqname[i], sqname[j]) == 0) {
		    wgt[j] = Wwgt[i];
		    }
		}
	}

	fclose(fp);
	free(sptr);
	for (i = 0; i < nseq; i++) {
	    free(Wsqname[i]);
	}
	free(beginIdx);
	free(endIdx);
	free(beginWgt);
	free(endWgt);
	free(Wwgt);
	free(Wsqname);
	free(tmpWwgt);
}
 

/* Function: print_all_weigths()
 * 
 * Purpose:  For each training sequence, print the weight is has
 *           been given by the weighting scheme
 *           
 * Args:     fp   - where to print the output (usu. stdout)
 *           msa  - alignment and associated info 
 *           
 * Return:   (void)                         
 */

void
print_all_weights(FILE *fp, MSA *msa)
{
  int idx;			/* counter for sequences */

  fputs("%Individual sequence weights:\n", fp);
				/* weight for each sequence */
  for (idx = 0; idx < msa->nseq; idx++) 
    {
      fprintf(fp, "%-12s %7.2f\n", 
	      msa->sqname[idx],
	      msa->wgt[idx]);

    }
  fputs("\n", fp);
}


/* Function: RelativeEntropy()
 * 
 * Date: March 03 marwis
 *
 * Purpose:  Calculate the relative entropy (relative to the background dist)
 *            of the emission probabilities in the match states  
 *           
 * Args:     hmm      - an hmm, logoddsified and ready to use  
 *           relative_entropy  - vector for the relative entropy at each model node
 *
 * Return:   relative_entropy
 *
 */

void
RelativeEntropy(struct plan7_s *hmm, float *relative_entropy)
{
    int k;			/* counter for model position */
    int x;			/* counter for symbols        */
    float vec[Alphabet_size];
    /* Symbol emission scores
     */
    
    for (k = 1; k <= hmm->M; k++) {
	/* match/insert emissions in main model */
	relative_entropy[k] = 0;
	for (x = 0; x < Alphabet_size; x++) {
	    relative_entropy[k] += hmm->mat[k][x]*sreLOG2(hmm->mat[k][x]/hmm->null[x]);
	}
    }
}

/* Function: RelativeEntropyVec()
 * 
 * Date: Nov 04 marwis
 *
 * Purpose:  Calculate the relative entropy of vec1 relative to vec2
 *            
 *           
 * Args:     vec1  - a vector
 *           vec2  - another vector (typically reference distribution)
 *
 * Return:   relative_entropy
 *
 */

float
RelativeEntropyVec(float *vec1, float *vec2)
{
    int x;			/* counter for symbols  */
    float relative_entropy;

    relative_entropy = 0;
    for (x = 0; x < Alphabet_size; x++) {
      relative_entropy += vec1[x]*sreLOG2(vec1[x]/vec2[x]);
    }
    return relative_entropy;
}


/* Function: SetEffSeqNum()
 * Date:     SRE, Fri Jan  2 10:56:00 1998 [StL]
 * 
 * Purpose:  Sets the effective sequence number from relative entropy calc. 
 *
 *           A background model Pb(x) is defined by "prioritizing" a null vector.
 *           Pm(x) is the posterior prob. to emit amino acid x in match state m.
 *           The relative entropy between the model and background is obtained by averaging
 *           over all match states of the model, i.e.
 *           rel_ent = 1/M * sum_M(sum_aa(Pm(x) * log(Pm(x)/Pb(x)))) 
 *
 *           Pm(x) depends on the effective sequence number. Given a target entropy, 
 *           an iterative procedure defines the effective sequence number so as to  
 *           match the target entropy.          
 * 
 * Args:     te      - target entropy per column
 *           hmm     - model. allocated, set with initial MAP parameters.
 *           dsq     - dealigned digitized seqs the model is based on
 *           prior   - prior distributions for parameterizing model
 *           tr      - array of traces for each sequence
 *           MSA     - the multiple seq. alignment
 *           
 * Return:   (void)
 *           hmm changed to an ME HMM
 *           ainfo changed, contains ME weights          
 */
void
SetEffSeqNum(float *te, struct plan7_s *hmm, char **dsq, struct p7prior_s *pri, 
		struct p7trace_s **tr, MSA *msa)
{

  float  grad;                  /* gradient                        */
  float  relative_entropy;      /* current best relative entropy   */
  float  new_entropy;           /* relative entropy at new weights */
  float  new_wgt;               /* new and temporary weights       */
  float  new_eff_nseq;          /* new effective seq weight        */
  int    idx;                   /* counter over HMM match states   */
  float  max_diff;              /* defining how close to the target we need to get */
  float  *nullvec;              /* null vector to prioritize with Dirichlet */
  float  eff_nseq;              /* effective number of seqs */
  float  converge_criterion;   
  float  target_entropy;        /* target entropy */

  /* Allocations and Initialisation   */

  target_entropy = *te;
  nullvec     = MallocOrDie (sizeof(float) * Alphabet_size);

  grad        = 1.0;
  max_diff    = 0.01;
  eff_nseq    = FSum(msa->wgt, msa->nseq); 
  new_entropy =  0;
  new_eff_nseq = eff_nseq;
  FSet(nullvec, Alphabet_size, 0.0); 

  /* make sure the HMM is ready */
  Plan7SWConfig(hmm, 0.5, 0.5);
  P7Logoddsify(hmm, TRUE);

  /* Priorify a vector of zero counts using the standard prior. This will be the 
     background distribution relative to which we compute the relative entropy of the HMM */
  P7PriorifyEmissionVector(nullvec, pri, pri->mnum, pri->mq, pri->m, NULL);

  /* Calculate the initial relative entropy using the initial effective sequence number */
  relative_entropy = 0;
  for (idx = 1; idx < hmm->M; idx++) 
    {
      relative_entropy += RelativeEntropyVec(hmm->mat[idx], nullvec);
    }
  relative_entropy = relative_entropy / hmm->M;
  
  printf("Initial effective seq. number:  %f  Initial entropy:  %f   Target_entropy:  %f\n\n", 
	 eff_nseq, relative_entropy, target_entropy);
  new_entropy = relative_entropy;

  /* iterate until the entropy is close enough to the target entropy */
  while (fabs(target_entropy - new_entropy) > max_diff) { 

    /* calculate the entropy using the current effective sequence number */

    /* Loop over the main nodes of the model */
    relative_entropy = 0;
    for (idx = 1; idx < hmm->M; idx++) 
      {
	relative_entropy += RelativeEntropyVec(hmm->mat[idx], nullvec);
      }
    relative_entropy = relative_entropy / hmm->M;
    
    /* update the current entropy */
    new_entropy = relative_entropy;
    
    /* Calculate a gradient in entropy space. This optimisation can't go wrong really,
       as long as we make sure the gradient is follws 0 < grad < 1 */
    grad = (target_entropy - new_entropy) / (new_entropy + target_entropy);
    new_eff_nseq = new_eff_nseq + new_eff_nseq*grad;
    
    printf("New effective seq. number:  %f  Current gradient:   %f   Current entropy:  %f\n", 
	   new_eff_nseq, grad, new_entropy);

    /* normalise and scale to update the weights to the new effective sequence number */
    FNorm(msa->wgt, msa->nseq);
    FScale(msa->wgt, msa->nseq, (float) new_eff_nseq);
    
    //    printf("Weight %f\n",hmm->mat[1]);
    /* re-fill the HMM with probs. using the saved trace and count but new sequence weights */
    ZeroPlan7(hmm);
    for (idx = 0; idx < msa->nseq; idx++)
      P7TraceCount(hmm, dsq[idx], msa->wgt[idx], tr[idx]);
    P7PriorifyHMM(hmm, pri);
    Plan7SWConfig(hmm, 0.5, 0.5);
    P7Logoddsify(hmm, TRUE);
  }

  free(nullvec);
}


/* Function: P7TransitionExpected()
 *
 * marwis April 2003
 *           
 * Purpose:  Combines Forward and Backward matrices into a probability
 *           matrix of transition probabilities. We're interested in the
 *           expected number of times a certain transition is used 
 *
 *           The 7 entries in column i of this matrix (trs) are the logs of the
 *           expected number of times each of the 7 transition occured between node i and j 
 *           The caller must allocate space for the matrix, or use an "old" allocated trs
 *           to build further on
 *                      
 * Args:     L        - length of sequence
 *           hmm      - the model
 *           forward  - pre-calculated forward matrix
 *           backward - pre-calculated backward matrix
 *           trs       - pre-allocated matrix
 *           
 * Return:   void
 */

void
P7TransitionExpected(int L, float wgt,
		     struct plan7_s *hmm,
		      struct dpmatrix_s *forward,
		     struct dpmatrix_s *backward,
		     int **ret_trs)
     
{
  int i;
  int k;
  int idx;
  int sc;
  int **trs;
  
  
  /* allocate a matrix for expected number of transitions [1..M][0..6], initialize all values to -INFTY
     ("worth" 0 in sum of logs) 
  */
  trs = MallocOrDie (sizeof(int *) * hmm->M);
  for (idx = 0; idx < hmm->M; idx++) 
    trs[idx] = MallocOrDie(sizeof(int) * 7);
  
  for (idx = 0; idx < hmm->M; idx++) {
    for (i = 0; i < 7; i++) {
      trs[idx][i] = -INFTY;
    }
  }

  sc = backward->xmx[0][XMN];

     /* Calculate n_k: the expected number of times transition k was used by the sequence.
	We're working with scaled log probs from the backward and forward matrices,
	so we have to make sure that what we do fits this. 
	(For example, we convert the sequence weights to fit)
      */
     /* Main state transition scores     [1.M-1][0.6]  
	Forward matrix: 0..L rows, 0..M-1 columns. Row 0 and Col 0 are -INFTY for mmx, imx, dmx
	Backward matrix: 0..L rows, 0..M-1 columns. 
	Both bck and fwd: mmx[i][k] = sum of probs up to state k + the i:th aa having been 
	emitted by m-state at node k
	imx[i][k] = sum of probs up to state k + the i:th aa having been emitted 
	by i-state at node k
	dmx[i][k]  = ILogsum(mmx[i][k-1] + hmm->tsc[k-1][TMD], dmx[i][k-1] + hmm->tsc[k-1][TDD]);
			  imx and dmx don't exist (= -INFTY) at node hmm->M
			  But, we don't care as these are wing-folded to -INFTY 
     */
  
  for (k = 1; k < hmm->M; k++) {  /* loop over all nodes having transitions leaving [1..M-1] */     
    for (i = 1; i <= L; i++)    /* loop over the length of the sequence */
      {      
	/* sum over all positions */
	/* mm: state 1->M-1 */
	trs[k][TMM]  = ILogsum((backward->mmx[i][k+1] + 
				forward->mmx[i-1][k] + hmm->tsc[TMM][k]), 
			       trs[k][TMM]);
	
	trs[k][TMI]  = ILogsum((backward->imx[i][k] +       /* mi: state 1->M-1 */
				forward->mmx[i-1][k] + hmm->tsc[TMI][k]),
			       trs[k][TMI]);
	//	if (k != hmm->M-1) {                                 /* md: state 1->M-2 */
	  trs[k][TMD]  = ILogsum((backward->dmx[i][k+1] +     /* having emitted i aa and then taken md */
				  forward->mmx[i][k] + hmm->tsc[TMD][k]),
				 trs[k][TMD]);
	  //	}
	
	trs[k][TIM]  = ILogsum((backward->mmx[i][k+1] +       /* im: state 1->M-1 */
				forward->imx[i-1][k] + hmm->tsc[TIM][k]),
			       trs[k][TIM]);
	
	trs[k][TII]  = ILogsum((backward->imx[i][k] +          /* ii: state 1->M-1 */     
				forward->imx[i-1][k] + hmm->tsc[TII][k]),
			       trs[k][TII]);
	//	if (k != 1) {                                          /* dm: state 2->M-1 */
	  trs[k][TDM]  = ILogsum((backward->mmx[i][k+1] +
				  forward->dmx[i-1][k] + hmm->tsc[TDM][k]),
				 trs[k][TDM]); 
	  //	}
	
	  //	if (k != 1 & k != hmm->M-1) {                          /* dd: state 2->M-2 */
	  trs[k][TDD]  = ILogsum((backward->dmx[i][k+1] +
				  forward->dmx[i][k] + hmm->tsc[TDD][k]),
				 trs[k][TDD]); 
	  //	}
      }	 
    trs[k][TMM] += -sc + INTSCALE*sreLOG2(wgt);
    trs[k][TMI] += -sc + INTSCALE*sreLOG2(wgt); 
    trs[k][TMD] += -sc + INTSCALE*sreLOG2(wgt);
    trs[k][TIM] += -sc + INTSCALE*sreLOG2(wgt);
    trs[k][TII] += -sc + INTSCALE*sreLOG2(wgt);
    trs[k][TDM] += -sc + INTSCALE*sreLOG2(wgt);
    trs[k][TDD] += -sc + INTSCALE*sreLOG2(wgt);
    ret_trs[k][TMM] = ILogsum(trs[k][TMM], ret_trs[k][TMM]);
    ret_trs[k][TMI] = ILogsum(trs[k][TMI], ret_trs[k][TMI]); 
    ret_trs[k][TMD] = ILogsum(trs[k][TMD], ret_trs[k][TMD]);
    ret_trs[k][TIM] = ILogsum(trs[k][TIM], ret_trs[k][TIM]);
    ret_trs[k][TII] = ILogsum(trs[k][TII], ret_trs[k][TII]);
    ret_trs[k][TDM] = ILogsum(trs[k][TDM], ret_trs[k][TDM]);
    ret_trs[k][TDD] = ILogsum(trs[k][TDD], ret_trs[k][TDD]);
    
  }
  for (idx = 0; idx < hmm->M; idx++) {free(trs[idx]);}
  free(trs);
}




/* Function: P7TraceCountNegatives()
 * 
 * Date: March 03 marwis
 *
 * Purpose:  Count a traceback into a count-based HMM structure.
 *           (Usually as part of a model parameter re-estimation.)
 *           
 * Args:     hmm   - counts-based HMM
 *           dsq   - digitized sequence that traceback aligns to the HMM (1..L)
 *           wt    - weight on the sequence
 *           tr    - alignment of seq to HMM
 *           t     - negative counts
 *
 * Return:   (void)
 */
void
P7TraceCountNegatives(float **t, struct plan7_s *hmm, char *dsq, float wt, struct p7trace_s *tr)
{
  int tpos;                     /* position in tr */
  int i;			/* symbol position in seq */
 
  for (tpos = 0; tpos < tr->tlen; tpos++)
    {
      i = tr->pos[tpos];

      /* Emission counts. 
       * Don't bother counting null states N,J,C.
       */
      /*     if (tr->statetype[tpos] == STM) 
	     P7CountSymbol(hmm->mat[tr->nodeidx[tpos]], dsq[i], wt);
	     else if (tr->statetype[tpos] == STI) 
	     P7CountSymbol(hmm->ins[tr->nodeidx[tpos]], dsq[i], wt); */
      
      /* State transition counts
       */
      switch (tr->statetype[tpos]) {
      case STS:
	break;			/* don't bother; P=1 */
      case STN:
	switch (tr->statetype[tpos+1]) {
	case STB: /*hmm->xt[XTN][MOVE] += wt;*/ break;   /* marwis, take away this later */
	case STN: /*hmm->xt[XTN][LOOP] += wt;*/ break;
	default:
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
      case STB:
	switch (tr->statetype[tpos+1]) {
	case STM:/* hmm->begin[tr->nodeidx[tpos+1]] += wt;*/ break;  /* marwis, take away this later */
	case STD:/* hmm->tbd1 += wt;                      */ break;
	default:      
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
      case STM:
	switch (tr->statetype[tpos+1]) {
	case STM: t[tr->nodeidx[tpos]][TMM] += wt; break;
	case STI: t[tr->nodeidx[tpos]][TMI] += wt; break;
	case STD: t[tr->nodeidx[tpos]][TMD] += wt; break;
	case STE: /*hmm->end[tr->nodeidx[tpos]]  += wt; */ break;
	default:    
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
      case STI:
	switch (tr->statetype[tpos+1]) {
	case STM: t[tr->nodeidx[tpos]][TIM] += wt; break;
	case STI: t[tr->nodeidx[tpos]][TII] += wt; break;
	default:    
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
      case STD:
	switch (tr->statetype[tpos+1]) {
	case STM: t[tr->nodeidx[tpos]][TDM] += wt; break;
	case STD: t[tr->nodeidx[tpos]][TDD] += wt; break;
	case STE: /* ignore; p(D->E) = 1.0 */           break;
	default: 
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
      case STE:
	switch (tr->statetype[tpos+1]) {
	case STC:/* hmm->xt[XTE][MOVE] += wt;*/ break;
	case STJ:/* hmm->xt[XTE][LOOP] += wt;*/ break;
	default:     
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
     case STJ:
	switch (tr->statetype[tpos+1]) {
	case STB:/* hmm->xt[XTJ][MOVE] += wt;*/ break;
	case STJ:/* hmm->xt[XTJ][LOOP] += wt;*/ break;
	default:     
	  Die("illegal state transition %s->%s in traceback", 
	      Statetype(tr->statetype[tpos]),
	      Statetype(tr->statetype[tpos+1]));
	}
	break;
      case STC:
	switch (tr->statetype[tpos+1]) {
	case STT:/* hmm->xt[XTC][MOVE] += wt;*/ break;
	case STC:/* hmm->xt[XTC][LOOP] += wt;*/ break;
	default:     
	 Die("illegal state transition %s->%s in traceback", 
	     Statetype(tr->statetype[tpos]),
	     Statetype(tr->statetype[tpos+1])); 
	}
	break;
      case STT:
	break;			/* T is the last. It makes no transitions. */
      default:
	Die("illegal state %s in traceback", 
	    Statetype(tr->statetype[tpos]));
      }
    }
}



/* Function: SequencesAlignment() 
 * 
 * Date:     March 2003, marwis
 *
 * Purpose:  Given an alignment, return unaligned
 *           sequence array. 
 *
 * Args:     msa      - alignment to sequence
 *           ret_seq - RETURN: array of sequences
 *           ret_sqlen - RETURN: sequence lengths
 *           
 *           
 */ 
void
SequencesAlignment(MSA *msa, char ***ret_seq, int **ret_sqlen)
{
  char   **seq;
  int    *sqlen;      /* array of sequence lengths */
  int    idx;			/* counter for sequences     */
  int    upos;			/* position in unaligned seq */
  int    apos;			/* position in aligned seq   */

  seq = (char **) MallocOrDie (sizeof(char *) * msa->nseq);
  sqlen = MallocOrDie (sizeof(int) * msa->nseq);

  /* Find lengths of each unaligned seq */
  for (idx = 0; idx < msa->nseq; idx++) {
      sqlen[idx] = 0;
      for (apos = 0; apos < msa->alen; apos++) {
	  if (!isgap(msa->aseq[idx][apos]))  /* skip gaps */
	      sqlen[idx]++;
      }
  }

  for (idx = 0; idx < msa->nseq; idx++) {
      seq[idx] = (char *) MallocOrDie (sizeof(char) * (sqlen[idx]));
      for (apos = 0, upos = 0; apos < msa->alen; apos++) {
	  if (! isgap(msa->aseq[idx][apos]))  /* skip gaps */
	      seq[idx][upos++] = msa->aseq[idx][apos];
      }
  }

  *ret_seq = seq;
  *ret_sqlen = sqlen;
}


/* Function: adjust_transitions_trueneg()
 * Date:     Sept 2003, marwis
 *
 * Note:     Similar to ATP, bur here true negative sequences are used.
 *
 * Purpose: A set of negative sequences is aligned to the HMM and the transition
 *          probabilities of the model are adjusted depending on how differences
 *          in the paths of positive and negatives sequences
 *
 *           1) Calculate measure of divergence between positive seqs paths
 *              and negative seqs paths through the model.
 *           Idea: Look at difference in distribution on D and M nodes
 *           Define RelEnt = D(pos||neg) + D(neg||pos) 
 *           Define eps    = 1-exp(-RelEnt)   (0->1 , 0 when pos and neg eq. dist.
 *           2) Adjust transitions
 *           Idea: Go along a gradient, away from negative behaviour, towards positive
 *           Approach: Notice positives contain more match states than negatives 
 *                     (or if it's vice versa)
 *           Let epsilon (or a derivative of epsilon) be incrementor
 *           Set T(xy, new) = (1 +-epsilon) * T(xy, old), where + or - is
 *           decided on above
 *           If pos and neg take similar paths -> eps=0 -> no change
 *
 * Args:     hmm     - model. allocated, set with initial MAP parameters.
 *           dsq     - dealigned digitized seqs the model is based on
 *           msa     - multiple alignment (the input to hmmbuild)
 *           tr      - array of fake traces for each positive sequence        
 *           nrseqs  - negative sequences
 *           nsqinfo - info about negative sequences
 *           nnum    - number of neg. seqs
 *           nnoise  - number of neg. seqs to be used in the model adjustment
 *           noseq   - TRUE if using same number of neg as pos
 *           maxnoseq - maximum number of positive seqs [1000]
 *
 * Return:   (void)
 *           hmm with adjusted transitions
 *        
 */
void
adjust_transitions_trueneg(struct plan7_s *hmm,  char **dsq, MSA *msa,
			   struct p7trace_s **tr, char **nrseqs, SQINFO *nsqinfo, 
			   int nnum, int nnoise, int noseq, int maxnoseq,
			   int max_iter)
{
  
  struct dpmatrix_s *mx;        /* DP matrix, growable                     */
  char  **useq;                 /* unaligned aseq's, (the seq's from the msa) */ 
  int   *usqlen;                /* lenghts of unaligned aseq's */   
  int    sqlen;                 /* seq. length of random seq. */
  struct p7trace_s **rndtrs;    /* tracers of generated seq's aligned to the HMM */ 
  struct seqscore_s *rnd;       /* keeps track of score, length etc for random seq's */
  float  *eps;                  /* divergence measure */  
  float  *relative_entropy;     /* stores the relative entropy (emission) for each model position (1..k) */
  float  *posscore, *new_posscore; /* pos. scores before and after iteration */
  float  score;
  
  float  *rndwgt;	        /* weights of random seqs (I use 1 for all) */
  float  wgtsum;                /* sum of msa->wgt, positive seqs' weights */
  struct state_densities *pos;  /* percentage of pos. transitions per column */
  struct state_densities *neg;  /* percentage of pos. transitions per column */
  
  float  disc;                  /* measure of discrimination */
  float  new_disc;              /* new  measure of discrimination*/
  float avpos, avneg;           /* average pos./neg. score */
  int    i1;                    /* number of iterations */
  float  converge_criterion;    /* cut-off abs((disc-new_disc)/new_disc) */ 
  int    i, idx;                /* indices */ 
  float    x;                   /* help param. */

  printf("nnoise: %d  nnum %d  nnoise %d    noseq %d   maxnoseq %d \n\n",
	 nnoise, nnum, nnoise, noseq, maxnoseq);
  if(noseq)
    nnoise = msa->nseq;        /* same number of noise seqs as positive seqs */
  if (nnoise > maxnoseq) 
    nnoise = maxnoseq;
  if (nnoise > nnum)
    nnoise = nnum;  
  
  /* logoddsify the model */ 
  P7Logoddsify(hmm, TRUE);
  
  /* get positive unaligned sequences */
  new_posscore = MallocOrDie (msa->nseq * sizeof(float));
  posscore = MallocOrDie (msa->nseq * sizeof(float));
  SequencesAlignment(msa, &useq, &usqlen);
  
  /* Make some allocations */
  pos = MallocOrDie (sizeof(struct state_densities));
  neg = MallocOrDie (sizeof(struct state_densities));
  pos->M = MallocOrDie (sizeof(float) * (hmm->M+1));
  pos->D = MallocOrDie (sizeof(float) * (hmm->M+1));
  pos->I = MallocOrDie (sizeof(float) * (hmm->M+1));
  neg->M = MallocOrDie (sizeof(float) * (hmm->M+1));
  neg->D = MallocOrDie (sizeof(float) * (hmm->M+1));
  neg->I = MallocOrDie (sizeof(float) * (hmm->M+1));
  
  /* Tracer structure for random seq's aligned to the HMM */
  rndtrs  = MallocOrDie(sizeof(struct p7trace_s *) * nnoise);
  rndwgt = MallocOrDie(sizeof(float) * nnoise);
  
  relative_entropy = MallocOrDie ((hmm->M+1) * sizeof(float));   
  eps              = MallocOrDie ((hmm->M+1) * sizeof(float));
  FSet(rndwgt, nnoise, 1.0);

  
  /* The negative sequences are scored and the nnoise best are
     kept for adjusting the HMM */
  NegativesScore(hmm, &rnd, nnoise, nnum, nrseqs, nsqinfo);
  
  /* Create a DP matrix; initially only two rows big, but growable;
   * we overalloc by 25 rows (L dimension) when we grow; not growable
   * in model dimension, since we know the hmm size
   */
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0); 
  
  /* Align the negative seqs. to the HMM. and collect traces */  
  for (i = 0; i < nnoise; i++)
    {
      rnd->dsq[i] = DigitizeSequence(nrseqs[i], rnd->sqinfo[i].len);
      if (P7ViterbiSize(rnd->sqinfo[i].len, hmm->M) <= RAMLIMIT)
	(void) P7Viterbi(rnd->dsq[i], rnd->sqinfo[i].len, hmm, mx, &(rndtrs[i]));
      else
	(void) P7SmallViterbi(rnd->dsq[i], rnd->sqinfo[i].len, hmm, mx, &(rndtrs[i]));
    }
  
  wgtsum = 0;
  /* Get scores for true members before adjusting the HMM */
  for (i = 0; i < msa->nseq; i++) {
    posscore[i] =  P7TraceScore(hmm, dsq[i], tr[i]);
    wgtsum += msa->wgt[i]; 
  }
  
  /* Calculate the discrimination pre adjusting HMM */
  avpos = avneg = 0 ;
  for (i = 0; i < msa->nseq; i++) {
    avpos += posscore[i]*msa->wgt[i];
  } 
  avpos /= wgtsum;
  
  for (i = 0; i < nnoise; i++) { 
    avneg += rnd->score[i]; 
  }
  avneg /= nnoise;
  disc = avpos - avneg;
  
  if (disc/hmm->M > 100000000) 
    printf("HMM already discriminative; left untouched\n");
  else {
    /* Steepest descents optimization;
     * iterate until discrimination positives|negatives converges.
     */
    printf("maxiter %d \n", max_iter);
    new_disc = 0 ;
    i1 = 0;
    while (++i1 <= max_iter) {  
      
      /* We have traces for the pos. and neg. sequences to the HMM 
	 Count the percentage of different types of transitions and
	 change the HMM where negatives diverge much from positives      
	 collect how the transitions are distributed on M, I and D 
	 pos->M[1..M], pos->D[1..M]  1..M is the nodeidx of the traceback 
	 The usage is weighted but divided by wgtsum (i.e. distributed 0 to 1) 
	 We add a Laplace prior to the counts to avoid 0
      */
      
      position_densities(hmm, dsq, msa->wgt, msa->nseq, tr, pos);
      position_densities(hmm, rnd->dsq, rndwgt, nnoise, rndtrs, neg);
      
      /* Calculate a measure of divergence (called eps) between the paths of pos. and neg. sequences 
	 Trying out using the distribution of paths through Delete and Match states at 
	 each node. Very ad hoc, of course... */
      
      /* Calculating the Rel. Ent using a pseudocount of 0.5 to avoid div. by zero 
	 [D(pos || neg) + D(neg || pos)] / 2
      */
      for (idx = 1; idx <= hmm->M; idx++) {
	relative_entropy[idx] =	
	  pos->M[idx]*sreLOG2(pos->M[idx]/neg->M[idx]) 
	  +  pos->D[idx]*sreLOG2(pos->D[idx]/neg->D[idx])
	  +  neg->M[idx]*sreLOG2(neg->M[idx]/pos->M[idx])
	  +  neg->D[idx]*sreLOG2(neg->D[idx]/pos->D[idx]);
	eps[idx] = 1 - exp(-relative_entropy[idx]/2);
      }
      
      // wmdgrad "default" (hmmbuild) if x=1 
      x=1;
      for (idx=2; idx <= hmm->M; idx++) { 
	if(pos->D[idx] < neg->D[idx]) { /* More match among pos, adjust towards match states */
	  hmm->t[idx-1][TMM] = (1 + x*eps[idx]) * hmm->t[idx-1][TMM]; /* pro match */
	  hmm->t[idx-1][TMD] = hmm->t[idx-1][TMD]/(1 + x*eps[idx]); /* anti match */
	  hmm->t[idx-1][TDM] = (1 + x*eps[idx]) * hmm->t[idx-1][TDM]; /* pro match */
	  hmm->t[idx-1][TDD] = hmm->t[idx-1][TDD]/(1 + x*eps[idx]); /* anti match */
	}
	else{ /* Less match among pos, adjust towards delete states */
	  hmm->t[idx-1][TMM] = hmm->t[idx-1][TMM]/(1 + x*eps[idx]); /* anti match */
	  hmm->t[idx-1][TMD] = (1 + x*eps[idx]) * hmm->t[idx-1][TMD]; /* pro delete */
	  hmm->t[idx-1][TDM] = hmm->t[idx-1][TDM]/(1 + x*eps[idx]); /* anti match */
	  hmm->t[idx-1][TDD] = (1 + x*eps[idx]) * hmm->t[idx-1][TDD]; /* pro delete */
	}
      }
      
      /* Convert count based HMM to probabilities */
      Plan7Renormalize(hmm);  
      Plan7LSConfig(hmm);
      P7Logoddsify(hmm, TRUE);
      
      /* Get new tracers (post adjusting the HMM) */  
      for (i=0; i < nnoise; i++) {
	if (P7ViterbiSize(rnd->sqinfo[i].len, hmm->M) <= RAMLIMIT)
	  score = P7Viterbi(rnd->dsq[i], rnd->sqinfo[i].len, hmm, mx, &(rndtrs[i]));
	else
	  score = P7SmallViterbi(rnd->dsq[i], rnd->sqinfo[i].len, hmm, mx, &(rndtrs[i]));
	
	rnd->newscore[i] = score;     
      }  
      
      /* get new positive scores */
      for (i = 0; i < msa->nseq; i++) {
	if (P7ViterbiSize(usqlen[i], hmm->M) <= RAMLIMIT)
	  score = P7Viterbi(dsq[i], usqlen[i], hmm, mx, &(tr[i]));
	else
	  score = P7SmallViterbi(dsq[i], usqlen[i], hmm, mx, &(tr[i]));
	  new_posscore[i] = score;
      }
	
      /* Calculate the discrimination as the average (posscore-negscore).
	 Scores are weighted */
	
      avpos = avneg = 0 ;
      for (i = 0; i < msa->nseq; i++) {
	avpos += new_posscore[i]*msa->wgt[i];
      } 
      avpos /= wgtsum;
      
      for (i = 0; i < nnoise; i++) { 
	avneg += rnd->newscore[i]; 
      }
      avneg /= nnoise;
      new_disc = avpos - avneg;
      
      
      /* print the new scores */
      printf("Negatives \n\n");
      for(i = 0; i< nnoise; i++) 
	printf("%d %f %f \n", i, rnd->score[i], rnd->newscore[i]);
      printf("\nPositives\n\n");
      for(i = 0; i < msa->nseq; i++) 
	printf("%f %f %f\n", posscore[i], new_posscore[i], msa->wgt[i]);
	printf("average score positives - negative %f %f difference %f \n", avpos, avneg,  new_disc);
	fflush(stdout);    
	
	converge_criterion = fabs((new_disc-disc)/disc);
	disc = new_disc;
	if (converge_criterion < 1e-2) {
	  printf("   -- converge_criterion %f  -- i1 %d\n", converge_criterion, i1 );
	  break;
	}
	printf("   -- converge_criterion %f  -- i1 %d\n", converge_criterion, i1 );
    }
  }
  if (i1 == max_iter) printf("   -- exceeded maximum iterations; giving up --\n");
  
  /* Clean up */   
  FreePlan7Matrix(mx);
  for (idx = 0; idx < msa->nseq; idx++) {free(useq[idx]);}
  free(useq);
  free(usqlen);   
  for (idx = 0; idx < nnoise; idx++) P7FreeTrace(rndtrs[idx]);
  free(rndtrs);
  free(eps);
  free(relative_entropy);
  free(posscore);
  free(new_posscore);
  FreeSeqscore_s(rnd, nnoise);
  FreeState_densities(pos);
  FreeState_densities(neg);    
  return;
}



/* Function: adjust_transitions_grad()
 * Date:     March 2003
 *
 * Note:     
 *
 * Purpose:  Generate random sequences and align them to the HMM. The ones scoring 
 *           the highest are considered to adjust the trans. probs. of the HMM.
 *           1) Calculate measure of divergence between positive seqs paths
 *              and negative seqs paths through the model.
 *           Idea: Look at difference in distribution on D and M nodes
 *           Define RelEnt = D(pos||neg) + D(neg||pos) 
 *           Define eps    = 1-exp(-RelEnt)   (0->1 , 0 when pos and neg eq. dist.
 *           2) Adjust transitions
 *           Idea: Go along a gradient, away from negative behaviour, towards positive
 *           Approach: Notice positives contain more match states than negatives 
 *                     (or if it's vice versa)
 *           Let epsilon (or a derivative of epsilon) be incrementor
 *           Set T(xy, new) = (1 +-epsilon) * T(xy, old), where + or - is
 *           decided on above
 *           If pos and neg take similar paths -> eps=0 -> no change
 *
 * Args:     hmm     - model. allocated, set with initial MAP parameters.
 *           dsq     - dealigned digitized seqs the model is based on
 *           msa     - multiple alignment (the input to hmmbuild)
 *           seed    - seed to generate random number
 *           lenmean  - mean length of random sequence (325 aa)
 *           lensd    - std dev of random seq length   (200)
 *           fixedlen - if nonzero, override lenmean, always this len
 *           nsample  - number of negative seqs to generate
 *           
 *           tr      - array of fake traces for each positive sequence        
 *           
 * Return:   (void)
 *           hmm with adjusted transitions
 *        
 */
void
adjust_transitions_grad(struct plan7_s *hmm,  char **dsq, MSA *msa,
		   struct p7trace_s **tr, int nsample, int nnoise,
		   int noseq, int maxnoseq, int max_iter)
{
  
  struct dpmatrix_s *mx;        /* DP matrix, growable                     */
  char  **useq;                 /* unaligned aseq's, (the seq's from the msa) */ 
  int   *usqlen;                /* lenghts of unaligned aseq's */   
  int    sqlen;                 /* seq. length of random seq. */
  struct p7trace_s **rndtrs;    /* tracers of generated seq's aligned to the HMM */ 
  struct seqscore_s *rnd;       /* keeps track of score, length etc for random seq's */
  float  *eps;                  /* divergence measure */  
  float  *relative_entropy;     /* stores the relative entropy (emission) for each model position (1..k) */
  float  *posscore, *new_posscore; /* pos. scores before and after iteration */
  float  score;

  float  *rndwgt;	        /* weights of random seqs (I use 1 for all) */
  float  wgtsum;                /* sum of msa->wgt, positive seqs' weights */
  struct state_densities *pos;  /* percentage of pos. transitions per column */
  struct state_densities *neg;  /* percentage of pos. transitions per column */

  float  disc;                  /* measure of discrimination */
  float  new_disc;              /* new  measure of discrimination*/
  float avpos, avneg;           /* average pos./neg. score */
  int    i1;                    /* number of iterations */
  float  converge_criterion;    /* cut-off abs((disc-new_disc)/new_disc) */ 
  float  lenmean;          	/* mean of Gaussian for random seq len */
  float  lensd;			/* s.d. of Gaussian for random seq len */
  int    seed;                  /* random number seed              */
  int     fixedlen;	        /* fixed length, or 0 if unused    */
  int    i, idx;                /* indices */ 

  /* These are the default values from hmmcalibrate, whatever -h says.. */
  lenmean = 325.;
  lensd = 200.;
  fixedlen = 0;
  seed     = (int) time ((time_t *) NULL);
  
 if(noseq)
    nnoise = msa->nseq;  /* same number of noise seqs as positive seqs */
  if (nnoise > maxnoseq) 
    nnoise = maxnoseq;
  printf("...sampling %d sequences, keeping the %d best, max %d iter \n", nsample, nnoise, max_iter);


  /* logoddsify the model */ 
  P7Logoddsify(hmm, TRUE);

  /* get positive unaligned sequences */
  new_posscore = MallocOrDie (msa->nseq * sizeof(float));
  posscore = MallocOrDie (msa->nseq * sizeof(float));
  SequencesAlignment(msa, &useq, &usqlen);

   /* Make some allocations */
   pos = MallocOrDie (sizeof(struct state_densities));
   neg = MallocOrDie (sizeof(struct state_densities));
   pos->M = MallocOrDie (sizeof(float) * (hmm->M+1));
   pos->D = MallocOrDie (sizeof(float) * (hmm->M+1));
   pos->I = MallocOrDie (sizeof(float) * (hmm->M+1));
   neg->M = MallocOrDie (sizeof(float) * (hmm->M+1));
   neg->D = MallocOrDie (sizeof(float) * (hmm->M+1));
   neg->I = MallocOrDie (sizeof(float) * (hmm->M+1));

   /* Tracer structure for random seq's aligned to the HMM */
   rndtrs  = MallocOrDie(sizeof(struct p7trace_s *) * nnoise);
   rndwgt = MallocOrDie(sizeof(float) * nnoise);

   relative_entropy = MallocOrDie ((hmm->M+1) * sizeof(float));   
   eps              = MallocOrDie ((hmm->M+1) * sizeof(float));
   FSet(rndwgt, nnoise, 1.0);

   /* Generate a number of random sequences and return the best scoring ones  
      in the structure rnd */
   SampleAndScore(hmm, &rnd, nsample, nnoise, seed, fixedlen, lenmean, lensd);
   
  /* Create a DP matrix; initially only two rows big, but growable;
   * we overalloc by 25 rows (L dimension) when we grow; not growable
   * in model dimension, since we know the hmm size
   */
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0); 
 
 /* Align the random seqs. to the HMM. and collect traces */  
   for (i = 0; i < nnoise; i++)
     {
       if (P7ViterbiSize(rnd->sqlen[i], hmm->M) <= RAMLIMIT)
	 (void) P7Viterbi(rnd->dsq[i], rnd->sqlen[i], hmm, mx, &(rndtrs[i]));
	else
	  (void) P7SmallViterbi(rnd->dsq[i], rnd->sqlen[i], hmm, mx, &(rndtrs[i]));
      }

   wgtsum = 0;
   /* Get scores for true members before adjusting the HMM */
   for (i = 0; i < msa->nseq; i++) {
     posscore[i] =  P7TraceScore(hmm, dsq[i], tr[i]);
     wgtsum += msa->wgt[i]; 
   }


   /* Calculate the discrimination pre adjusting HMM */
   avpos = avneg = 0 ;
   for (i = 0; i < msa->nseq; i++) {
     avpos += posscore[i]*msa->wgt[i];
   } 
   avpos /= wgtsum;
   
   for (i = 0; i < nnoise; i++) { 
     avneg += rnd->score[i]; 
   }
   avneg /= nnoise;
   disc = avpos - avneg;

   if (disc/hmm->M > 100000000) 
     printf("HMM already discriminative; left untouched\n");
   else {
     printf("iter avg-pos-sc avg-neg-sc diff convergence\n");
     printf("---- ---------- ---------- ---- -----------\n");
     printf("%4d %8.1f %8.1f %9.1f %5s\n",
         0, avpos, avneg, disc, "-");
     
     /* Steepest descents optimization;
      * iterate until discrimination positives|negatives converges.
      */
     new_disc = 0 ;
     i1 = 0;
     while (++i1 <= max_iter) {  
       
     /* We have traces for the pos. and neg. sequences to the HMM 
	Count the percentage of different types of transitions and
	change the HMM where negatives diverge much from positives      
	collect how the transitions are distributed on M, I and D 
	pos->M[1..M], pos->D[1..M]  1..M is the nodeidx of the traceback 
       The usage is weighted but divided by wgtsum (i.e. distributed 0 to 1) 
       We add a Laplace prior to the counts to avoid 0
    */
     
     position_densities(hmm, dsq, msa->wgt, msa->nseq, tr, pos);
     position_densities(hmm, rnd->dsq, rndwgt, nnoise, rndtrs, neg);
     
     /* Calculate a measure of divergence (called eps) between the paths of pos. and neg. sequences 
	Trying out using the distribution of paths through Delete and Match states at 
	each node. Very ad hoc, of course... */
     
     /* Calculating the Rel. Ent using a pseudocount of 0.5 to avoid div. by zero 
       [D(pos || neg) + D(neg || pos)] / 2
     */
     for (idx = 1; idx <= hmm->M; idx++) {
       relative_entropy[idx] =	
	 pos->M[idx]*sreLOG2(pos->M[idx]/neg->M[idx]) 
	 +  pos->D[idx]*sreLOG2(pos->D[idx]/neg->D[idx])
	 +  neg->M[idx]*sreLOG2(neg->M[idx]/pos->M[idx])
	 +  neg->D[idx]*sreLOG2(neg->D[idx]/pos->D[idx]);
       eps[idx] = 1 - exp(-relative_entropy[idx]/2);
     }
     
     for (idx=2; idx <= hmm->M; idx++) { 
       if(pos->D[idx] < neg->D[idx]) { /* More match among pos, adjust towards match states */
      	 hmm->t[idx-1][TMM] = (1 + eps[idx]) * hmm->t[idx-1][TMM]; /* pro match */
      	 hmm->t[idx-1][TMD] = hmm->t[idx-1][TMD]/(1 + eps[idx]); /* anti match */
      	 hmm->t[idx-1][TDM] = (1 + eps[idx]) * hmm->t[idx-1][TDM]; /* pro match */
      	 hmm->t[idx-1][TDD] = hmm->t[idx-1][TDD]/(1 + eps[idx]); /* anti match */
       }
       else{ /* Less match among pos, adjust towards delete states */
      	 hmm->t[idx-1][TMM] = hmm->t[idx-1][TMM]/(1 + eps[idx]); /* anti match */
      	 hmm->t[idx-1][TMD] = (1 + eps[idx]) * hmm->t[idx-1][TMD]; /* pro delete */
     	 hmm->t[idx-1][TDM] = hmm->t[idx-1][TDM]/(1 + eps[idx]); /* anti match */
      	 hmm->t[idx-1][TDD] = (1 + eps[idx]) * hmm->t[idx-1][TDD]; /* pro delete */
      }
     }

     
     /* Convert count based HMM to probabilities */
     Plan7Renormalize(hmm);  
     Plan7LSConfig(hmm);
     P7Logoddsify(hmm, TRUE);
    
    /* Get new tracers (post adjusting the HMM) */  
    for (i=0; i < nnoise; i++) {
      if (P7ViterbiSize(rnd->sqlen[i], hmm->M) <= RAMLIMIT)
	score = P7Viterbi(rnd->dsq[i], rnd->sqlen[i], hmm, mx, &(rndtrs[i]));
      else
	score = P7SmallViterbi(rnd->dsq[i], rnd->sqlen[i], hmm, mx, &(rndtrs[i]));
      rnd->newscore[i] = score;     
    }  
    
    /* get new positive scores */
    for (i = 0; i < msa->nseq; i++) {
      if (P7ViterbiSize(usqlen[i], hmm->M) <= RAMLIMIT)
	score = P7Viterbi(dsq[i], usqlen[i], hmm, mx, &(tr[i]));
      else
	score = P7SmallViterbi(dsq[i], usqlen[i], hmm, mx, &(tr[i]));
      new_posscore[i] = score;
    }
           
    /* Calculate the discrimination as the average (posscore-negscore).
       Scores are weighted */

    avpos = avneg = 0 ;
    for (i = 0; i < msa->nseq; i++) {
      avpos += new_posscore[i]*msa->wgt[i];
    } 
     avpos /= wgtsum;
     
     for (i = 0; i < nnoise; i++) { 
       avneg += rnd->newscore[i]; 
     }
     avneg /= nnoise;
     new_disc = avpos - avneg;
          
     converge_criterion = fabs((new_disc-disc)/disc);
     disc = new_disc;
     printf("%4d %8.1f %8.1f %9.1f %7.3f\n",
         i1, avpos, avneg, disc, converge_criterion);
     if (converge_criterion < 1e-2) {
       break;
     }
   }
   }
   if (i1 == max_iter) printf("   -- exceeded maximum iterations; giving up --\n");

    /* Clean up */   
   FreePlan7Matrix(mx);
   for (idx = 0; idx < msa->nseq; idx++) {free(useq[idx]);}
   free(useq);
   free(usqlen);   
   for (idx = 0; idx < nnoise; idx++) P7FreeTrace(rndtrs[idx]);
   free(rndtrs);
   free(eps);
   free(relative_entropy);
   free(posscore);
   free(new_posscore);
   FreeSeqscore_s(rnd, nnoise);
   FreeState_densities(pos);
   FreeState_densities(neg);    
   return;
}

/* Function: adjust_transitions_post()
 * Date:     marwis
 *
 * Note:     
 *
 * Purpose:  Generate random sequences and align them to the HMM. The ones scoring 
 *           the highest are considered to adjust the trans. probs. of the HMM.
 *           1) Calculate measure of divergence between positive seqs paths
 *              and negative seqs paths through the model.
 *           Idea: Look at difference in distribution on D and M nodes
 *           Define RelEnt = D(pos||neg) + D(neg||pos) 
 *           Define eps    = 1-exp(-RelEnt)   (0->1 , 0 when pos and neg eq. dist.
 *           2) Adjust transitions
 *           Idea: Go along a gradient, away from negative behaviour, towards positive
 *           Approach: Notice positives contain more match states than negatives 
 *                     (or if it's vice versa)
 *           Let epsilon (or a derivative of epsilon) be incrementor
 *           Set T(xy, new) = (1 +-epsilon) * T(xy, old), where + or - is
 *           decided on above
 *           If pos and neg take similar paths -> eps=0 -> no change
 *
 * Args:     hmm     - model. allocated, set with initial MAP parameters.
 *           dsq     - dealigned digitized seqs the model is based on
 *           msa     - multiple alignment (the input to hmmbuild)
 *           seed    - seed to generate random number
 *           lenmean  - mean length of random sequence (325 aa)
 *           lensd    - std dev of random seq length   (200)
 *           fixedlen - if nonzero, override lenmean, always this len
 *           nsample  - number of negative seqs to generate
 *           
 *           tr      - array of fake traces for each positive sequence        
 *           
 * Return:   (void)
 *           hmm with adjusted transitions
 *        
 */
void
adjust_transitions_post(struct plan7_s *hmm,  char **dsq, MSA *msa,
		   struct p7trace_s **tr, int nsample, int nnoise,
		   int noseq, int maxnoseq)
{
  char  *seq;
  char  **useq;         /* unaligned aseq's              */ 
  int   *usqlen;        /* lenghts of unaligned aseq's */   
  char  *rnddsq;
  struct p7trace_s **rndtrs;   
  int   *seqlen; 
  struct seqscore_s *rnd;  
  struct seqscore_s *new_rnd;
  float  *rndwgt;	    /* weights to assign to alignment          */
  float  *pos_score;
  float  *eps;              /* divergence measure */  
  float cij[8];             /* transit counts vector */
  int   j;                  /* next index in alignment */
  int   nodeidx;            /* nodeindex for the HMM */
  
  float  *posscore;
  float  *new_posscore;
  float  score;
  int    sqlen;                /* seq. length of random seq. */
  int    i, k, idx; 
  char   *rndname;
  float  disc;                /* measure of discrimnation */
  float  new_disc;            /* new  measure of discrimination*/
  float avpos, avneg, origavpos;
  int    i1;                  /* number of iterations */
  int    max_iter;            /* max. number of iterations */
  float  converge_criterion;  /* cut-off abs((disc-new_disc)/disc) */
  float  wgtsum;              /* sum of msa->wgt */
 float  lenmean;          	/* mean of Gaussian for random seq len */
  float  lensd;			/* s.d. of Gaussian for random seq len */
  int    seed;                /* random number seed              */
  int     fixedlen;	      /* fixed length, or 0 if unused    */
  struct dpmatrix_s *mx;        /* DP matrix, growable */
  struct dpmatrix_s *fwd;
  struct dpmatrix_s *bck;
  int  **trs, **trsneg;             /* expected number of times a transition was used */
  float **grad;               /* gradient for transitions */  
 
  SQINFO *rnd_sqinfo;
    /* List of indices of scores: high to low */
  

  /* These are the default values from hmmcalibrate, whatever documentation says.. */
  lenmean = 325.;
  lensd = 200.;
  fixedlen = 0;
  seed     = (int) time ((time_t *) NULL);
  max_iter = 1;

  if(noseq)
    nnoise = msa->nseq;  /* same number of noise seqs as positive seqs */
  if (nnoise > maxnoseq) 
    nnoise = maxnoseq;  
  printf("...sampling %d sequences, keeping the %d best, max %d iter \n", nsample, nnoise, max_iter);

  rndname = MallocOrDie (sizeof(char *));
  
  /* logoddsify the model */ 
  P7Logoddsify(hmm, TRUE);
  
  /* Allocate stuff to deal with scores */

  new_posscore = MallocOrDie (msa->nseq * sizeof(float));
  posscore = MallocOrDie (msa->nseq * sizeof(float));

  /* some allocations */
  trs = MallocOrDie (sizeof(int *) * hmm->M);
  trsneg = MallocOrDie (sizeof(int *) * hmm->M);
  rndtrs  = MallocOrDie(sizeof(struct p7trace_s *) * nnoise);
  rndwgt = MallocOrDie(sizeof(float) * nnoise);
  FSet(rndwgt, nnoise, 1.0);

  grad = MallocOrDie (sizeof(float *) * hmm->M);
  for (idx = 0; idx < hmm->M; idx++) {
      trs[idx] = MallocOrDie (sizeof(int) * 7);
      trsneg[idx] = MallocOrDie (sizeof(int) * 7);
      grad[idx] = MallocOrDie (sizeof(float) * 7);
  }

  /* Get positive unaligned sequences */
  SequencesAlignment(msa, &useq, &usqlen);

  wgtsum = 0;
   /* Get scores for positive seqs before adjusting the HMM */
   for (i = 0; i < msa->nseq; i++) {
     posscore[i] =  P7TraceScore(hmm, dsq[i], tr[i]);
     wgtsum += msa->wgt[i]; 
   }
 
  /* Generate a number of random sequences and return the best scoring ones 
     in the structure rnd */
  SampleAndScore(hmm, &rnd, nsample, nnoise, seed, fixedlen, lenmean, lensd);
  

   /* Calculate the discrimination pre adjusting HMM */
   avpos = avneg = 0 ;
   for (i = 0; i < msa->nseq; i++) {
     avpos += posscore[i]*msa->wgt[i];
   } 
   avpos /= wgtsum;
   
   for (i = 0; i < nnoise; i++) { 
     avneg += rnd->score[i]; 
   }
   avneg /= nnoise;
   disc = avpos - avneg;
   origavpos = avpos;
  
   printf("iter avg-pos-sc avg-neg-sc diff convergence\n");
   printf("---- ---------- ---------- ---- -----------\n");
   printf("%4d %8.1f %8.1f %9.1f %5s\n",
	  0, avpos, avneg, disc, "-");

 /* Create a DP matrix; initially only two rows big, but growable;
   * we overalloc by 25 rows (L dimension) when we grow; not growable
   * in model dimension, since we know the hmm size
   */
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0); 

  /* Steepest descents optimization;
   * iterate until discrimination positives|negatives converges.
   */
  new_disc = 0 ;
  i1 = 0;
  while (++i1 <= max_iter) {  
    
    /* initialise/reset all values in the matrices to -INFTY or 0
       (-INFTY "worth" 0 in sum of logs) 
    */

    for (idx = 0; idx < hmm->M; idx++) {
      for (i = 0; i < 7; i++) {
	trs[idx][i] = -INFTY;
	trsneg[idx][i] = -INFTY;
	grad[idx][i] = 0;
      }
    }
     
    /* For each sequence,.... get backward and forward matrix,.....
     * calculate expected number of times each transition is used
     * over all seqs     
     * 
     */
    for (i = 0; i < msa->nseq; i++) {
      score = P7Forward(dsq[i], usqlen[i], hmm, &fwd);
      score = P7Backward(dsq[i], usqlen[i], hmm, &bck);
        (void) P7TransitionExpected(usqlen[i], msa->wgt[i], hmm, fwd, bck, trs);
      FreePlan7Matrix(fwd);
      FreePlan7Matrix(bck);	 
    }
    
    /* For each random sequence,.... get backward and forward matrix,.....
     * calculate expected number of times each transition is used summed
     * over all seqs
     * 
     */
    for (i = 0; i < nnoise; i++) {
      score = P7Forward (rnd->dsq[i], rnd->sqlen[i], hmm, &fwd);
      score = P7Backward(rnd->dsq[i], rnd->sqlen[i], hmm, &bck);
      (void) P7TransitionExpected(rnd->sqlen[i], rndwgt[i], hmm, fwd, bck, trsneg);
      FreePlan7Matrix(fwd);
      FreePlan7Matrix(bck);
  }
  
    for (k = 1; k < hmm->M; k++)
      for (i = 0; i < 7; i++) 
	grad[k][i] =  (Score2Prob(trs[k][i], 1.)/wgtsum -  Score2Prob(trsneg[k][i], 1.)/nnoise);
    
    /* take the step along the gradient,...how far? */
    
    for (k = 1; k < hmm->M; k++) 
      for (i = 0; i < 7; i++) {
	while (grad[k][i] < -1) {grad[k][i] /= 2;}   /* security check to avoid zero probs, rarely (if ever (?)) used it seems*/
	     hmm->t[k][i] = (1 + grad[k][i]) * hmm->t[k][i];
	     if (hmm->t[k][i] < 0) {printf("Below k %d i %d value %f\n",  k, i, hmm->t[k][i]);}
      }
    
      /* Convert count based HMM to probabilities */
      Plan7Renormalize(hmm);  
      Plan7LSConfig(hmm);
      P7Logoddsify(hmm, TRUE);
      
      /* get new negative scores */
      for (idx=0; idx < nnoise; idx++) {
	rnddsq = DigitizeSequence(rnd->seq[idx], rnd->sqlen[idx]);
	
	if (P7ViterbiSize(rnd->sqlen[idx], hmm->M) <= RAMLIMIT)
	  score = P7Viterbi(rnddsq, rnd->sqlen[idx], hmm, mx, &(rndtrs[i]));
	else
	  score = P7SmallViterbi(rnddsq, rnd->sqlen[idx], hmm, mx, &(rndtrs[i]));       
	rnd->newscore[idx] = score;     
	free(rnddsq);
      }  
      
     /* get new positive scores, and keep the tracer updated */
      for (i = 0; i < msa->nseq; i++) {
	if (P7ViterbiSize(usqlen[i], hmm->M) <= RAMLIMIT)
	  score = P7Viterbi(dsq[i], usqlen[i], hmm, mx, &(tr[i]));
	else
	  score = P7SmallViterbi(dsq[i], usqlen[i], hmm,  mx, &(tr[i]));
	new_posscore[i] = score;
      }
             
     /* Calculate the discrimination as the average (posscore-negscore).
	For now it's weighted, not sure this is what we want */
     avpos = avneg = 0 ;
     for (i = 0; i < msa->nseq; i++) {
	 avpos += new_posscore[i]*msa->wgt[i];
     } 
     avpos /= wgtsum;

     for (i = 0; i < nnoise; i++) { 
	 avneg += rnd->newscore[i]; 
     }
     avneg /= nnoise;
     new_disc = avpos - avneg;

     converge_criterion = fabs((disc-new_disc)/disc);
     disc = new_disc;
     printf("%4d %8.1f %8.1f %9.1f %7.3f\n",
         i1, avpos, avneg, disc, converge_criterion);

     if (converge_criterion < 1e-2) {printf("   -- converged --\n"); break;}
     if (avpos/origavpos < 0.99)    {printf("   -- iteration stopped; positives dropped > 1 percent in score\n"); break;} 
     
  }
  if (i1 == max_iter) printf("   -- exceeded maximum iterations; giving up --\n");   
     
    /* Clean up */  
     FreePlan7Matrix(mx); 
     free(rndtrs);
     for (idx = 0; idx < hmm->M; idx++) {free(grad[idx]);}
     free(grad);
     for (idx = 0; idx < hmm->M; idx++) {free(trsneg[idx]);}
     free(trsneg);
     for (idx = 0; idx < msa->nseq; idx++) {free(useq[idx]);}
     free(useq);
     free(usqlen);   
     free(posscore);
     free(new_posscore);
     free(rndname);
 
    return;
}


/* Function: FreeSeqscore_s()    
*/

void
FreeSeqscore_s(struct seqscore_s *rnd, int dim1)
{
  int idx;
  free(rnd->score);
  if (rnd->newscore != NULL) free(rnd->newscore);
  free(rnd->queue);
  free(rnd->sqlen);
  for (idx = 0; idx < dim1; idx++) {free(rnd->dsq[idx]);}
  for (idx = 0; idx < dim1; idx++) {free(rnd->seq[idx]);}
  if (rnd->sqinfo != NULL) free(rnd->sqinfo);
  free(rnd);
}

/* Function: FreeState_densitiies()
 */

void
FreeState_densities(struct state_densities *dens)
{
  if (dens->M != NULL) free(dens->M);
  if (dens->D != NULL) free(dens->D);
  if (dens->I != NULL) free(dens->I);
  free(dens);
}

/* Function: position_densities()
 * Date:     marwis March 2003
 * 
 * Purpose:  Calculate use of M,D,I from tracebacks, keeping them
 *           in a position specific array. The final array
 *           is normalized position-specifically too, according
 *           to how many sequences contributed data to this
 *           position. 
 *           Very much ad hoc
 *           
 *           Code related to (derived from) position_average_score.
 *           
 * Args:     hmm       - HMM structure, scores valid
 *           dsq       - digitized unaligned sequences
 *           wgt       - weights on the sequences
 *           nseq      - number of sequences
 *           tr        - array of nseq tracebacks that aligns each dsq to hmm
 *           dens   - RETURN: [0]1..M arrays of M,I,D use percentage
 *           ret_avg   - RETURN: overall average full-length, one-domain score
 *           
 * Return:   1 on success, 0 on failure.          
 *           pernode is malloc'ed [0]1..M by CALLER and filled here.
 */
void
position_densities(struct plan7_s    *hmm, 
		   char             **dsq, 
		   float             *wgt,
		   int                nseq,
		   struct p7trace_s **tr,
		   struct state_densities *dens)
{
  int    pos;                   /* position in seq */
  int    sym;
  int    tpos;                  /* position in trace/state sequence */
  float *counts;                /* counts at each position */
  float  avg;                   /* RETURN: average overall */
  int    k;                     /* counter for model position */
  int    idx;                   /* counter for sequence number */
  
  /* Allocations
   */
  counts = MallocOrDie ((hmm->M+1) * sizeof(float));
  FSet(dens->M, hmm->M+1, 0.);
  FSet(dens->I, hmm->M+1, 0.);
  FSet(dens->D, hmm->M+1, 0.);
  
  FSet(counts,  hmm->M+1, 0.);
    
  /* Loop over traces, accumulate weighted usage per position
   */
  for (idx = 0; idx < nseq; idx++)
    for (tpos = 0; tpos < tr[idx]->tlen; tpos++)
      {
	pos = tr[idx]->pos[tpos];
	sym = (int) dsq[idx][tr[idx]->pos[tpos]];
	k   = tr[idx]->nodeidx[tpos];
	
	/* Counts: how many times did we use this model position 1..M?
         * (weighted)
	 */
	if (tr[idx]->statetype[tpos] == STM || tr[idx]->statetype[tpos] == STD)
	  counts[k] += wgt[idx];

	/* How many times did we go through M, D, I, respectively.
	 */
	if (tr[idx]->statetype[tpos] == STM) 
	  dens->M[k] += wgt[idx];
	else if (tr[idx]->statetype[tpos] == STD) 
	  dens->D[k] += wgt[idx];
	else if (tr[idx]->statetype[tpos] == STI) 
	  dens->I[k] += wgt[idx];
      }
  /* Divide accumulated paths for each state by all paths. Add laplace prior
   */
  
    for (k = 1; k <= hmm->M; k++)
    {
	dens->M[k] = (dens->M[k] + 0.5) / (counts[k]+1);
	dens->D[k] = (dens->D[k] + 0.5) / (counts[k]+1);
	dens->I[k] = (dens->I[k] ) / (counts[k]+1);
    }
  
  free(counts);
  return;
}


/* Function: SampleAndScore()
 * 
 * Purpose:  Sample and score a given number of noise sequences
 *           
 *           
 * 
 * Args:     hmm      - an hmm, logoddsified and ready to use  
 *           ret_rnd  - structure for generated seqs to return
 *           seed     - random number seed
 *           nsample  - number of seqs to synthesize
 *           lenmean  - mean length of random sequence
 *           lensd    - std dev of random seq length
 *           fixedlen - if nonzero, override lenmean, always this len
 *           nnoise    - number of noise sequences to return
 * 
 * Return:   rnd structure (ret_rnd);
 *           Caller responsible for freeing rnd
 */

void
SampleAndScore(struct plan7_s *hmm, struct seqscore_s **ret_rnd, int nsample, int nnoise, 
	       int seed, int fixedlen, float lenmean, float lensd)
{
  float  randomseq[MAXABET];
  float  p1;
  int    sqlen;      /* seq. length of random seq. */
  int    laps;       /* laps in main loops */
  int    added;      /* number of random seqs added to list */ 
  float  lowscore;   /* lowest scoring seq. in list */
  float  score;      /* score of akigned seq */  
  char  *rnddsq;     /* digitised random seq */
  char  *seq;        /* sequence being generated */ 
  struct seqscore_s *rnd;  /* structure for generated seqs */
  int    i;          /* index */
  struct dpmatrix_s *mx;        /* DP matrix, growable                     */

  /* Allocation of rnd structure */
   rnd = MallocOrDie (sizeof(struct seqscore_s));  
  rnd->score = MallocOrDie (nnoise * sizeof(float));
  rnd->queue = MallocOrDie (nnoise * sizeof(int)); 
  rnd->sqlen = MallocOrDie (nnoise *sizeof(int));
  rnd->dsq = MallocOrDie (nnoise * sizeof(char *));
  rnd->seq = MallocOrDie (nnoise * sizeof(char *));
  rnd->newscore = MallocOrDie (nnoise * sizeof(float));
  rnd->sqinfo = NULL;
  added = 0;

  lowscore = FLT_MAX;
  sre_srandom(seed);
  P7DefaultNullModel(randomseq, &p1);

    /* Create a DP matrix; initially only two rows big, but growable;
   * we overalloc by 25 rows (L dimension) when we grow; not growable
   * in model dimension, since we know the hmm size
   * Is being Resized in Viterbi
   */
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0); 

  for (laps = 0; laps < nsample; laps++) {
      /* choose length of random sequence */
      if (fixedlen) sqlen = fixedlen;
      else do sqlen = (int) Gaussrandom(lenmean, lensd); while (sqlen < 1);
      /* generate it */
      seq = RandomSequence(Alphabet, randomseq, Alphabet_size, sqlen);
      rnddsq = DigitizeSequence(seq, sqlen);

      if (P7ViterbiSize(sqlen, hmm->M) <= RAMLIMIT)
	  score = P7Viterbi(rnddsq, sqlen, hmm, mx, NULL);
      else
	  score = P7SmallViterbi(rnddsq, sqlen, hmm, mx, NULL);
     
      /* store the first nnoise random sequences */
      if(added < nnoise) {
	  AddToList(rnd, score, sqlen, seq, rnddsq, added); 
	  added++;
	  
	  if (added == nnoise)  /*the list is full and we sort it */
	      {	      
		  /* rnd->queue contains indices to the scores in rnd->score: high to low */
		  IndexSortList(rnd->queue, rnd->score, nnoise);
		  lowscore = rnd->score[rnd->queue[nnoise-1]];
	      }
      }
      
      /* a new hit? */
      else if (score > lowscore) 
	  {
	      /* Push the list of seq.scores and place the new hit on the right place */
	      PushList(rnd, score, sqlen, seq, rnddsq, nnoise);
	      lowscore = rnd->score[rnd->queue[nnoise-1]];
	  }  
      
      free(rnddsq); 
      free(seq);
  }
    FreePlan7Matrix(mx);
    *ret_rnd = rnd;
  return;
}


/* Function: NegativesScore()
 * 
 * Purpose:  Sample and score a given number of noise sequences
 *           
 *           
 * 
 * Args:     hmm      - an hmm, logoddsified and ready to use  
 *           ret_rnd  - structure for generated seqs to return
 *           nnum     - number of negative sequences in alignment
 *           nrseq    - negative sequences
 *           nsqinfo  - info about negative sequences 
 *           nnoise   - number of noise sequences to return
 * 
 * Return:   rnd structure (ret_rnd);
 *           Caller responsible for freeing rnd
 */

void
NegativesScore(struct plan7_s *hmm, struct seqscore_s **ret_rnd, int nnoise,
	       int nnum, char **nrseq,  SQINFO *nsqinfo)
{
  float  randomseq[MAXABET];
  float  p1;
  int    sqlen;      /* seq. length of random seq. */
  int    laps;       /* laps in main loops */
  int    added;      /* number of random seqs added to list */ 
  float  lowscore;   /* lowest scoring seq. in list */
  float  score;      /* score of akigned seq */  
  char  *rnddsq;     /* digitised negative seq */
  struct seqscore_s *rnd;  /* structure for generated seqs */
  int    i;          /* index */
  struct dpmatrix_s *mx;        /* DP matrix, growable                     */

  /* Allocation of rnd structure */
  rnd = MallocOrDie (sizeof(struct seqscore_s));  
  rnd->score = MallocOrDie (nnoise * sizeof(float));
  rnd->queue = MallocOrDie (nnoise * sizeof(int)); 
  rnd->sqlen = MallocOrDie (nnoise *sizeof(int));
  rnd->dsq = MallocOrDie (nnoise * sizeof(char *));
  rnd->seq = MallocOrDie (nnoise * sizeof(char *));
  rnd->newscore = MallocOrDie (nnoise * sizeof(float));
  rnd->sqinfo = nsqinfo;
  added = 0;

  lowscore = FLT_MAX;
  P7DefaultNullModel(randomseq, &p1);

    /* Create a DP matrix; initially only two rows big, but growable;
   * we overalloc by 25 rows (L dimension) when we grow; not growable
   * in model dimension, since we know the hmm size
   * Is being Resized in Viterbi
   */
  mx = CreatePlan7Matrix(1, hmm->M, 25, 0); 

  for (laps = 0; laps < nnum; laps++) {
    rnddsq = DigitizeSequence(nrseq[laps], nsqinfo[laps].len);
      if (P7ViterbiSize(nsqinfo[laps].len, hmm->M) <= RAMLIMIT)
	  score = P7Viterbi(rnddsq, nsqinfo[laps].len, hmm, mx, NULL);
      else
	  score = P7SmallViterbi(rnddsq, nsqinfo[laps].len, hmm, mx, NULL);
     
      /* store the first negative sequences */
      if(added < nnoise) {
	  AddToList(rnd, score, nsqinfo[laps].len, nrseq[laps], rnddsq, added); 
	  added++;
	  
	  if (added == nnoise)  /*the list is full and we sort it */
	      {	      
		  /* rnd->queue contains indices to the scores in rnd->score: high to low */
		  IndexSortList(rnd->queue, rnd->score, nnoise);
		  lowscore = rnd->score[rnd->queue[nnoise-1]];
	      }
      }
      
      /* a new hit? */
      else if (score > lowscore) 
	  {
	      /* Push the list of seq.scores and place the new hit on the right place */
	      PushList(rnd, score, nsqinfo[laps].len, nrseq[laps], rnddsq, nnoise);
	      lowscore = rnd->score[rnd->queue[nnoise-1]];
	  }  
      
      free(rnddsq); 
  }
  FreePlan7Matrix(mx);
  *ret_rnd = rnd;
  return;
}


/* Function  PushList()
 * Date:     March 03  marwis
 *
 * Purpose:  Push seqscore list and add new sequence with attributes 
 *           
 *
 * Args:     struct seqscore_s *rnd;
 *           float score;
 *           int   sqlen
 *           char  *seq
 *           char  *dsq 
 *           int   nnoise 
 *
 * Return:   void
 *           Returns the list. The worst scoring seq. has been removed and 
 *           the new hit been ranked by its score. Values scoring lower have
 *           been pushed down the rank.
 */

void
PushList(struct seqscore_s *rnd, float score, int sqlen, char *seq, char *dsq, int nnoise)
{
  int i, k; 
  int idx;
  int lastidx;
  
  k = 0;
  //  printf("Score %f\n", score);
  while (score <= rnd->score[rnd->queue[k]]) {
    k++;
  }
  /* k is now the index to be given to the new seq. */
  
  /* Seqs scoring worse than the new seq are pushed down the list. */
  idx = nnoise-1;
  lastidx = rnd->queue[nnoise-1];  /* We need to remember the index of the lowest scoring seq. because
				    we are placing the new seq. on its memory location. */
  
  while (idx > k){ 
    rnd->queue[idx] = rnd->queue[idx-1]; idx--;
  }
  
  /* Erase the worst scoring sequence */
  /* Replace with the new sequence */
  
  rnd->queue[k] = lastidx; 
  
  rnd->dsq[lastidx] = ReallocOrDie(rnd->dsq[lastidx], sizeof(char *) * (sqlen +2));
  rnd->seq[lastidx] = ReallocOrDie(rnd->seq[lastidx], sizeof(char *) * sqlen); 
  rnd->score[lastidx] = score;    
  rnd->sqlen[lastidx] = sqlen;

  rnd->dsq[lastidx][0] = dsq[0];       /* as in DigitiseSequence */
  rnd->dsq[lastidx][sqlen+1] = dsq[sqlen+1];
  
  for (i = 0; i < sqlen; i++) {
    rnd->seq[lastidx][i] = seq[i];
    rnd->dsq[lastidx][i+1] = dsq[i+1];
  }
  
  return; 
}


/* Function  AddToList()
 * Date:     March 03 marwis
 *
 * Purpose:  Add to list
 *           
 * Args:     struct  seqscore_s *rnd;
 *           float   score;
 *           int     sqlen
 *           char    *seq
 *           char    *dsq 
 *           int     added 
 *
 * Return:   void
 *           Returns the seq.score list with a new member. Unsorted.
 */


void
AddToList(struct seqscore_s *rnd, float score, int sqlen, char *seq, char *dsq, int added)
{
  int i;

  rnd->seq[added] =  MallocOrDie (sqlen * sizeof(char *));
  rnd->dsq[added] =  MallocOrDie ((sqlen + 2) * sizeof(char *));   //doing as in DigitiseSequence

  rnd->score[added] = score;    
  rnd->sqlen[added] = sqlen;
  rnd->dsq[added][0] = dsq[0];       /* as in DigitiseSequence */
  rnd->dsq[added][sqlen+1] = dsq[sqlen+1];
  for (i = 0; i < sqlen; i++) {
    rnd->seq[added][i] = seq[i];
    rnd->dsq[added][i+1] = dsq[i+1];	  
  }
  return;
}


/* Function  IndexSortList()
 * Date:     March 2003 marwis
 *
 * Purpose:  Given a list of scores, a new list is created with the indices
 *           of the scores in falling order.
 *
 * Args:     int *ret_list;
 *           float *score;
 *           int   lth; 
 *
 * Return:   void
 *           Returns the sorted list of indices.
 */
   
 void
IndexSortList(int *ret_list, float *score, int lth)
{
  int   idx;
  float max;
  float min;
  int i, j;

  max = FLT_MAX;
  min = -FLT_MAX;
  

  /* Find highest score strictly lower than max */
  for (i = 0; i < lth; i++) 
    {
      for (j = 0; j < lth; j++) 
	{
	  if (score[j] < max & score[j] > min )
	    {
	      min = score[j];
	      idx = j;
	    }
	}
      ret_list[i] = idx;
      max = min;
      min = -FLT_MAX;
    }
  return;
}
